;; It was proposed by Christian Goldbach that every odd composite
;; number can be written as the sum of a prime and twice a square.

;; 9 = 7 + 2x1^2

;; 15 = 7 + 2x2^2

;; 21 = 3 + 2x3^2

;; 25 = 7 + 2x3^2

;; 27 = 19 + 2x2^2

;; 33 = 31 + 2x1^2


;; It turns out that the conjecture was false.

;; What is the smallest odd composite that cannot be written as the sum of a prime and twice a square?

(defun memoize (function)
  (let ((res-table (make-hash-table :test #'equal :size 100000)))
    (lambda (&rest args)
      (multiple-value-bind 
	    (value exists) 
	  (gethash args res-table)
	(when (not exists)
	  (setf value (apply function args))
	  (setf (gethash args res-table) value))
	value))))

(defun inner-prime? (num)
  (cond ((= 1 num) nil)
	((= 2 num) t)
	((= 0 (mod num 2)) nil)
	(t
	 (loop for fact = (ceiling (sqrt num))
	       then (if (oddp fact)
			(- fact 2)
			(- fact 1))
	       while (>= fact 2)
	       do
	       (when (= 0 (mod num fact))
		 (return-from inner-prime? nil)))
	 t)))

'(mapcar 
  #'car
  (remove-if-not
   #'identity
   (loop for ii from 1 to 100 collect (list ii (inner-prime? ii)))
   :key #'second))


(defvar *inner-prime?-mem* (memoize #'inner-prime?))

(defun prime? (num)
  (funcall *inner-prime?-mem* num))


;; (mapcar 
;;  #'car
;;  (remove-if-not
;;   #'identity
;;   (loop for ii from 1 to 100 collect (list ii (prime? ii)))
;;   :key #'second))

(defun make-prime-gen ()
  (let ((num 1))
    (lambda ()
      (if (oddp num)
	  (incf num 2)
	  (incf num))
      (loop while (not (prime? num))
	    finally (return num)
	    do
	    (incf num)))))

(loop with pgen = (make-prime-gen)
      repeat 10
      do
      (format t "~a~&" (funcall pgen)))

(defun passes-test (num)
  (loop with pgen = (make-prime-gen)
	for prime = (funcall pgen) then (funcall pgen)
	while (< prime num)
	for rem = (- num prime) then (- num prime)
	for sq = (sqrt (/ rem 2))
	do
	(when (= (floor sq) sq)
	  ;;(format t "passes-test: num~a prime:~a rem:~a sq:~a~&"
		;;  num prime rem sq)
	  (return-from passes-test t)))
  nil)

(time
 (loop for ii = 9 then (+ ii 2)
       do
       (when (and (not (prime? ii))
		  (not (passes-test ii)))
	 (format t "found: ii:~a~&" ii)
	 (return ii))))